@@ -10,44 +10,44 @@
 #include <stdio.h> /* for FILE* */
 #include "nsDebug.h"
 #include "nsTArray.h"
+#include "mozilla/FunctionTypeTraits.h"
 #include "mozilla/RefPtr.h"
 #include "mozilla/ReverseIterator.h"
 
-#if defined(DEBUG) || defined(MOZ_DUMP_PAINTING)
+#if defined(DEBUG) || defined(MOZ_DUMP_PAINTING) || defined(MOZ_LAYOUT_DEBUGGER)
 // DEBUG_FRAME_DUMP enables nsIFrame::List and related methods.
 // You can also define this in a non-DEBUG build if you need frame dumps.
-#define DEBUG_FRAME_DUMP 1
+#  define DEBUG_FRAME_DUMP 1
 #endif
 
 class nsContainerFrame;
 class nsIContent;
 class nsIFrame;
-class nsIPresShell;
 class nsPresContext;
 
 namespace mozilla {
+class PresShell;
 namespace layout {
 class FrameChildList;
 enum FrameChildListID {
   // The individual concrete child lists.
-  kPrincipalList = 0x1,
-  kPopupList = 0x2,
-  kCaptionList = 0x4,
-  kColGroupList = 0x8,
-  kSelectPopupList = 0x10,
-  kAbsoluteList = 0x20,
-  kFixedList = 0x40,
-  kOverflowList = 0x80,
-  kOverflowContainersList = 0x100,
-  kExcessOverflowContainersList = 0x200,
-  kOverflowOutOfFlowList = 0x400,
-  kFloatList = 0x800,
-  kBulletList = 0x1000,
-  kPushedFloatsList = 0x2000,
-  kBackdropList = 0x4000,
+  kPrincipalList,
+  kPopupList,
+  kCaptionList,
+  kColGroupList,
+  kAbsoluteList,
+  kFixedList,
+  kOverflowList,
+  kOverflowContainersList,
+  kExcessOverflowContainersList,
+  kOverflowOutOfFlowList,
+  kFloatList,
+  kBulletList,
+  kPushedFloatsList,
+  kBackdropList,
   // A special alias for kPrincipalList that suppress the reflow request that
   // is normally done when manipulating child lists.
-  kNoReflowPrincipalList = 0x8000
+  kNoReflowPrincipalList,
 };
 
 // A helper class for nsIFrame::Destroy[From].  It's defined here because
@@ -56,14 +56,10 @@ struct PostFrameDestroyData {
   PostFrameDestroyData(const PostFrameDestroyData&) = delete;
   PostFrameDestroyData() = default;
 
-  AutoTArray<RefPtr<nsIContent>, 50> mAnonymousContent;
-  AutoTArray<RefPtr<nsIContent>, 50> mGeneratedContent;
+  AutoTArray<RefPtr<nsIContent>, 100> mAnonymousContent;
   void AddAnonymousContent(already_AddRefed<nsIContent>&& aContent) {
     mAnonymousContent.AppendElement(aContent);
   }
-  void AddGeneratedContent(already_AddRefed<nsIContent>&& aContent) {
-    mGeneratedContent.AppendElement(aContent);
-  }
 };
 }  // namespace layout
 }  // namespace mozilla
@@ -83,19 +79,38 @@ class nsFrameList {
     VerifyList();
   }
 
-  nsFrameList(const nsFrameList& aOther)
-      : mFirstChild(aOther.mFirstChild), mLastChild(aOther.mLastChild) {}
+  // XXX: Ideally, copy constructor should be removed because a frame should be
+  // owned by one list.
+  nsFrameList(const nsFrameList& aOther) = default;
+
+  // XXX: ideally, copy assignment should be removed because we should use move
+  // assignment to transfer the ownership.
+  nsFrameList& operator=(const nsFrameList& aOther) = default;
+
+  /**
+   * Move the frames in aOther to this list. aOther becomes empty after this
+   * operation.
+   */
+  nsFrameList(nsFrameList&& aOther)
+      : mFirstChild(aOther.mFirstChild), mLastChild(aOther.mLastChild) {
+    aOther.Clear();
+    VerifyList();
+  }
+  nsFrameList& operator=(nsFrameList&& aOther) {
+    SetFrames(aOther);
+    return *this;
+  }
 
   /**
    * Infallibly allocate a nsFrameList from the shell arena.
    */
-  void* operator new(size_t sz, nsIPresShell* aPresShell);
+  void* operator new(size_t sz, mozilla::PresShell* aPresShell);
 
   /**
    * Deallocate this list that was allocated from the shell arena.
    * The list is required to be empty.
    */
-  void Delete(nsIPresShell* aPresShell);
+  void Delete(mozilla::PresShell* aPresShell);
 
   /**
    * For each frame in this list: remove it from the list then call
@@ -116,7 +131,7 @@ class nsFrameList {
   void SetFrames(nsIFrame* aFrameList);
 
   void SetFrames(nsFrameList& aFrameList) {
-    NS_PRECONDITION(!mFirstChild, "Losing frames");
+    MOZ_ASSERT(!mFirstChild, "Losing frames");
 
     mFirstChild = aFrameList.FirstChild();
     mLastChild = aFrameList.LastChild();
@@ -223,6 +238,32 @@ class nsFrameList {
 
   class FrameLinkEnumerator;
 
+  /**
+   * Split this list just before the first frame that matches aPredicate,
+   * and return a nsFrameList containing all the frames before it. The
+   * matched frame and all frames after it stay in this list. If no matched
+   * frame exists, all the frames are drained into the returned list, and
+   * this list ends up empty.
+   *
+   * aPredicate should be of this function signature: bool(nsIFrame*).
+   */
+  template <typename Predicate>
+  nsFrameList Split(Predicate&& aPredicate) {
+    static_assert(
+        std::is_same<
+            typename mozilla::FunctionTypeTraits<Predicate>::ReturnType,
+            bool>::value &&
+            mozilla::FunctionTypeTraits<Predicate>::arity == 1 &&
+            std::is_same<typename mozilla::FunctionTypeTraits<
+                             Predicate>::template ParameterType<0>,
+                         nsIFrame*>::value,
+        "aPredicate should be of this function signature: bool(nsIFrame*)");
+
+    FrameLinkEnumerator link(*this);
+    link.Find(aPredicate);
+    return ExtractHead(link);
+  }
+
   /**
    * Split this frame list such that all the frames before the link pointed to
    * by aLink end up in the returned list, while the remaining frames stay in
@@ -248,12 +289,18 @@ class nsFrameList {
 
   bool NotEmpty() const { return nullptr != mFirstChild; }
 
+  /**
+   * Return true if aFrame is on this list.
+   * @note this method has O(n) time complexity over the length of the list
+   * XXXmats: ideally, we should make this function #ifdef DEBUG
+   */
   bool ContainsFrame(const nsIFrame* aFrame) const;
 
   /**
    * Get the number of frames in this list. Note that currently the
    * implementation has O(n) time complexity. Do not call it repeatedly in hot
    * code.
+   * XXXmats: ideally, we should make this function #ifdef DEBUG
    */
   int32_t GetLength() const;
 
@@ -330,14 +377,7 @@ class nsFrameList {
           mEnd(aEnd) {
     }
 
-    Slice(const Slice& aOther)
-        :
-#ifdef DEBUG
-          mList(aOther.mList),
-#endif
-          mStart(aOther.mStart),
-          mEnd(aOther.mEnd) {
-    }
+    Slice(const Slice& aOther) = default;
 
    private:
 #ifdef DEBUG
@@ -359,14 +399,7 @@ class nsFrameList {
           mEnd(aSlice.mEnd) {
     }
 
-    Enumerator(const Enumerator& aOther)
-        :
-#ifdef DEBUG
-          mSlice(aOther.mSlice),
-#endif
-          mFrame(aOther.mFrame),
-          mEnd(aOther.mEnd) {
-    }
+    Enumerator(const Enumerator& aOther) = default;
 
     bool AtEnd() const {
       // Can't just check mEnd, because some table code goes and destroys the
@@ -385,7 +418,7 @@ class nsFrameList {
      * iterator that is at end!
      */
     nsIFrame* get() const {
-      NS_PRECONDITION(!AtEnd(), "Enumerator is at end");
+      MOZ_ASSERT(!AtEnd(), "Enumerator is at end");
       return mFrame;
     }
 
@@ -437,8 +470,7 @@ class nsFrameList {
     explicit FrameLinkEnumerator(const nsFrameList& aList)
         : Enumerator(aList), mPrev(nullptr) {}
 
-    FrameLinkEnumerator(const FrameLinkEnumerator& aOther)
-        : Enumerator(aOther), mPrev(aOther.mPrev) {}
+    FrameLinkEnumerator(const FrameLinkEnumerator& aOther) = default;
 
     /* This constructor needs to know about nsIFrame, and nsIFrame will need to
        know about nsFrameList methods, so in order to inline this put
@@ -446,13 +478,26 @@ class nsFrameList {
     inline FrameLinkEnumerator(const nsFrameList& aList, nsIFrame* aPrevFrame);
 
     void operator=(const FrameLinkEnumerator& aOther) {
-      NS_PRECONDITION(&List() == &aOther.List(), "Different lists?");
+      MOZ_ASSERT(&List() == &aOther.List(), "Different lists?");
       mFrame = aOther.mFrame;
       mPrev = aOther.mPrev;
     }
 
     inline void Next();
 
+    /**
+     * Find the first frame from the current position that satisfies
+     * aPredicate, and stop at it. If no such frame exists, then this method
+     * advances to the end of the list.
+     *
+     * aPredicate should be of this function signature: bool(nsIFrame*).
+     *
+     * Note: Find() needs to see the definition of Next(), so put this
+     * definition in nsIFrame.h.
+     */
+    template <typename Predicate>
+    inline void Find(Predicate&& aPredicate);
+
     bool AtEnd() const { return Enumerator::AtEnd(); }
 
     nsIFrame* PrevFrame() const { return mPrev; }
@@ -464,11 +509,22 @@ class nsFrameList {
 
   class Iterator {
    public:
+    // It is disputable whether these type definitions are correct, since
+    // operator* doesn't return a reference at all. Also, the iterator_category
+    // can be at most std::input_iterator_tag (rather than
+    // std::bidrectional_iterator_tag, as it might seem), because it is a
+    // stashing iterator. See also, e.g.,
+    // https://stackoverflow.com/questions/50909701/what-should-be-iterator-category-for-a-stashing-iterator
+    using value_type = nsIFrame* const;
+    using pointer = value_type*;
+    using reference = value_type&;
+    using difference_type = ptrdiff_t;
+    using iterator_category = std::input_iterator_tag;
+
     Iterator(const nsFrameList& aList, nsIFrame* aCurrent)
         : mList(aList), mCurrent(aCurrent) {}
 
-    Iterator(const Iterator& aOther)
-        : mList(aOther.mList), mCurrent(aOther.mCurrent) {}
+    Iterator(const Iterator& aOther) = default;
 
     nsIFrame* operator*() const { return mCurrent; }
 
@@ -547,14 +603,13 @@ inline bool operator!=(const nsFrameList::Iterator& aIter1,
 }
 
 namespace mozilla {
-namespace layout {
 
 /**
  * Simple "auto_ptr" for nsFrameLists allocated from the shell arena.
  * The frame list given to the constructor will be deallocated (if non-null)
  * in the destructor.  The frame list must then be empty.
  */
-class AutoFrameListPtr {
+class MOZ_RAII AutoFrameListPtr final {
  public:
   AutoFrameListPtr(nsPresContext* aPresContext, nsFrameList* aFrameList)
       : mPresContext(aPresContext), mFrameList(aFrameList) {}
@@ -575,12 +630,11 @@ union AlignedFrameListBytes {
 extern const AlignedFrameListBytes gEmptyFrameListBytes;
 }  // namespace detail
 
-}  // namespace layout
 }  // namespace mozilla
 
 /* static */ inline const nsFrameList& nsFrameList::EmptyList() {
   return *reinterpret_cast<const nsFrameList*>(
-      &mozilla::layout::detail::gEmptyFrameListBytes);
+      &mozilla::detail::gEmptyFrameListBytes);
 }
 
 #endif /* nsFrameList_h___ */
