@@ -5,13 +5,12 @@
 //! Specified color values.
 
 use super::AllowQuirks;
-#[cfg(feature = "gecko")]
-use crate::gecko_bindings::structs::nscolor;
 use crate::parser::{Parse, ParserContext};
 use crate::values::computed::{Color as ComputedColor, Context, ToComputedValue};
-use crate::values::generics::color::{GenericColorOrAuto, GenericCaretColor};
+use crate::values::generics::color::{GenericCaretColor, GenericColorOrAuto};
 use crate::values::specified::calc::CalcNode;
 use crate::values::specified::Percentage;
+use crate::values::CustomIdent;
 use cssparser::{AngleOrNumber, Color as CSSParserColor, Parser, Token, RGBA};
 use cssparser::{BasicParseErrorKind, NumberOrPercentage, ParseErrorKind};
 use itoa;
@@ -20,17 +19,51 @@ use std::io::Write as IoWrite;
 use style_traits::{CssType, CssWriter, KeywordsCollectFn, ParseError, StyleParseErrorKind};
 use style_traits::{SpecifiedValueInfo, ToCss, ValueParseErrorKind};
 
+/// A color space as defined in [1].
+///
+/// [1]: https://drafts.csswg.org/css-color-5/#typedef-colorspace
+#[derive(Clone, Copy, Debug, Eq, MallocSizeOf, Parse, PartialEq, ToCss, ToShmem)]
+pub enum ColorSpaceKind {
+    /// The sRGB color space.
+    Srgb,
+    /// The CIEXYZ color space.
+    Xyz,
+    /// The CIELAB color space.
+    Lab,
+    /// The CIELAB color space, expressed in cylindrical coordinates.
+    Lch,
+}
+
+/// A hue adjuster as defined in [1].
+///
+/// [1]: https://drafts.csswg.org/css-color-5/#typedef-hue-adjuster
+#[derive(Clone, Copy, Debug, Eq, MallocSizeOf, Parse, PartialEq, ToCss, ToShmem)]
+pub enum HueAdjuster {
+    /// The "shorter" angle adjustment.
+    Shorter,
+    /// The "longer" angle adjustment.
+    Longer,
+    /// The "increasing" angle adjustment.
+    Increasing,
+    /// The "decreasing" angle adjustment.
+    Decreasing,
+    /// The "specified" angle adjustment.
+    Specified,
+}
+
 /// A restricted version of the css `color-mix()` function, which only supports
-/// percentages and sRGB color-space interpolation.
+/// percentages.
 ///
 /// https://drafts.csswg.org/css-color-5/#color-mix
 #[derive(Clone, Debug, MallocSizeOf, PartialEq, ToShmem)]
 #[allow(missing_docs)]
 pub struct ColorMix {
+    pub color_space: ColorSpaceKind,
     pub left: Color,
     pub left_percentage: Percentage,
     pub right: Color,
     pub right_percentage: Percentage,
+    pub hue_adjuster: HueAdjuster,
 }
 
 // NOTE(emilio): Syntax is still a bit in-flux, since [1] doesn't seem
@@ -49,6 +82,9 @@ impl Parse for ColorMix {
             return Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError));
         }
 
+        let color_spaces_enabled = context.chrome_rules_enabled() ||
+            static_prefs::pref!("layout.css.color-mix.color-spaces.enabled");
+
         input.expect_function_matching("color-mix")?;
 
         // NOTE(emilio): This implements the syntax described here for now,
@@ -57,12 +93,18 @@ impl Parse for ColorMix {
         // https://github.com/w3c/csswg-drafts/issues/6066#issuecomment-789836765
         input.parse_nested_block(|input| {
             input.expect_ident_matching("in")?;
-            // TODO: support multiple interpolation spaces.
-            input.expect_ident_matching("srgb")?;
+            let color_space = if color_spaces_enabled {
+                ColorSpaceKind::parse(input)?
+            } else {
+                input.expect_ident_matching("srgb")?;
+                ColorSpaceKind::Srgb
+            };
             input.expect_comma()?;
 
             let left = Color::parse(context, input)?;
-            let left_percentage = input.try_parse(|input| Percentage::parse(context, input)).ok();
+            let left_percentage = input
+                .try_parse(|input| Percentage::parse(context, input))
+                .ok();
 
             input.expect_comma()?;
 
@@ -75,11 +117,18 @@ impl Parse for ColorMix {
 
             let left_percentage =
                 left_percentage.unwrap_or_else(|| Percentage::new(1.0 - right_percentage.get()));
+
+            let hue_adjuster = input
+                .try_parse(|input| HueAdjuster::parse(input))
+                .unwrap_or(HueAdjuster::Shorter);
+
             Ok(ColorMix {
+                color_space,
                 left,
                 left_percentage,
                 right,
                 right_percentage,
+                hue_adjuster,
             })
         })
     }
@@ -103,7 +152,9 @@ impl ToCss for ColorMix {
             (1.0 - percent.get() - other.get()).abs() <= f32::EPSILON
         }
 
-        dest.write_str("color-mix(in srgb, ")?;
+        dest.write_str("color-mix(in ")?;
+        self.color_space.to_css(dest)?;
+        dest.write_str(", ")?;
         self.left.to_css(dest)?;
         if !can_omit(&self.left_percentage, &self.right_percentage, true) {
             dest.write_str(" ")?;
@@ -115,24 +166,16 @@ impl ToCss for ColorMix {
             dest.write_str(" ")?;
             self.right_percentage.to_css(dest)?;
         }
+
+        if self.hue_adjuster != HueAdjuster::Shorter {
+            dest.write_str(" ")?;
+            self.hue_adjuster.to_css(dest)?;
+        }
+
         dest.write_str(")")
     }
 }
 
-/// The color scheme for a specific system color.
-#[cfg(feature = "gecko")]
-#[derive(Clone, Copy, Debug, MallocSizeOf, Parse, PartialEq, ToCss, ToShmem)]
-#[repr(u8)]
-pub enum SystemColorScheme {
-    /// The default color-scheme for the document.
-    #[css(skip)]
-    Default,
-    /// A light color scheme.
-    Light,
-    /// A dark color scheme.
-    Dark,
-}
-
 /// Specified color value
 #[derive(Clone, Debug, MallocSizeOf, PartialEq, ToShmem)]
 pub enum Color {
@@ -147,10 +190,9 @@ pub enum Color {
     },
     /// A complex color value from computed value
     Complex(ComputedColor),
-    /// Either a system color, or a `-moz-system-color(<system-color>, light|dark)`
-    /// function which allows chrome code to choose between color schemes.
+    /// A system color.
     #[cfg(feature = "gecko")]
-    System(SystemColor, SystemColorScheme),
+    System(SystemColor),
     /// A color mix.
     ColorMix(Box<ColorMix>),
     /// Quirksmode-only rule for inheriting color from the body
@@ -158,40 +200,25 @@ pub enum Color {
     InheritFromBodyQuirk,
 }
 
-/// System colors.
+/// System colors. A bunch of these are ad-hoc, others come from Windows:
+///
+///   https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getsyscolor
+///
+/// Others are HTML/CSS specific. Spec is:
+///
+///   https://drafts.csswg.org/css-color/#css-system-colors
+///   https://drafts.csswg.org/css-color/#deprecated-system-colors
 #[allow(missing_docs)]
 #[cfg(feature = "gecko")]
 #[derive(Clone, Copy, Debug, MallocSizeOf, Parse, PartialEq, ToCss, ToShmem)]
 #[repr(u8)]
 pub enum SystemColor {
-    #[css(skip)]
-    WindowBackground,
-    #[css(skip)]
-    WindowForeground,
-    #[css(skip)]
-    WidgetBackground,
-    #[css(skip)]
-    WidgetForeground,
-    #[css(skip)]
-    WidgetSelectBackground,
-    #[css(skip)]
-    WidgetSelectForeground,
-    #[css(skip)]
-    Widget3DHighlight,
-    #[css(skip)]
-    Widget3DShadow,
-    #[css(skip)]
-    TextBackground,
-    #[css(skip)]
-    TextForeground,
     #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
-    TextSelectBackground,
-    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
-    TextSelectForeground,
-    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
-    TextSelectBackgroundDisabled,
-    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
-    TextSelectBackgroundAttention,
+    TextSelectDisabledBackground,
+    #[css(skip)]
+    TextSelectAttentionBackground,
+    #[css(skip)]
+    TextSelectAttentionForeground,
     #[css(skip)]
     TextHighlightBackground,
     #[css(skip)]
@@ -235,6 +262,7 @@ pub enum SystemColor {
     #[css(skip)]
     ThemedScrollbarThumbInactive,
     Activeborder,
+    /// Background in the (active) titlebar.
     Activecaption,
     Appworkspace,
     Background,
@@ -242,16 +270,23 @@ pub enum SystemColor {
     Buttonhighlight,
     Buttonshadow,
     Buttontext,
+    /// Text color in the (active) titlebar.
     Captiontext,
     #[parse(aliases = "-moz-field")]
     Field,
+    /// Used for disabled field backgrounds.
+    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
+    MozDisabledfield,
     #[parse(aliases = "-moz-fieldtext")]
     Fieldtext,
+
     Graytext,
     Highlight,
     Highlighttext,
     Inactiveborder,
+    /// Background in the (inactive) titlebar.
     Inactivecaption,
+    /// Text color in the (inactive) titlebar.
     Inactivecaptiontext,
     Infobackground,
     Infotext,
@@ -276,13 +311,15 @@ pub enum SystemColor {
     /// Used to highlight valid regions to drop something onto.
     MozDragtargetzone,
     /// Used for selected but not focused cell backgrounds.
+    #[parse(aliases = "-moz-html-cellhighlight")]
     MozCellhighlight,
     /// Used for selected but not focused cell text.
+    #[parse(aliases = "-moz-html-cellhighlighttext")]
     MozCellhighlighttext,
-    /// Used for selected but not focused html cell backgrounds.
-    MozHtmlCellhighlight,
-    /// Used for selected but not focused html cell text.
-    MozHtmlCellhighlighttext,
+    /// Used for selected and focused html cell backgrounds.
+    Selecteditem,
+    /// Used for selected and focused html cell text.
+    Selecteditemtext,
     /// Used to button text background when hovered.
     MozButtonhoverface,
     /// Used to button text color when hovered.
@@ -303,10 +340,16 @@ pub enum SystemColor {
 
     /// Used for button text when pressed.
     #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
-    MozGtkButtonactivetext,
+    MozButtonactivetext,
+
+    /// Used for button background when pressed.
+    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
+    MozButtonactiveface,
+
+    /// Used for button background when disabled.
+    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
+    MozButtondisabledface,
 
-    /// Used for button text when pressed.
-    MozMacButtonactivetext,
     /// Background color of chrome toolbars in active windows.
     MozMacChromeActive,
     /// Background color of chrome toolbars in inactive windows.
@@ -348,6 +391,10 @@ pub enum SystemColor {
     #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
     MozAccentColorForeground,
 
+    /// The background-color for :autofill-ed inputs.
+    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
+    MozAutofillBackground,
+
     /// Media rebar text.
     MozWinMediatext,
     /// Communications rebar text.
@@ -361,6 +408,10 @@ pub enum SystemColor {
     /// colors.
     MozNativehyperlinktext,
 
+    /// As above, but visited link color.
+    #[css(skip)]
+    MozNativevisitedhyperlinktext,
+
     #[parse(aliases = "-moz-hyperlinktext")]
     Linktext,
     #[parse(aliases = "-moz-activehyperlinktext")]
@@ -378,14 +429,6 @@ pub enum SystemColor {
     #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
     MozColheaderhovertext,
 
-    /// Color of text in the (active) titlebar.
-    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
-    MozGtkTitlebarText,
-
-    /// Color of text in the (inactive) titlebar.
-    #[parse(condition = "ParserContext::in_ua_or_chrome_sheet")]
-    MozGtkTitlebarInactiveText,
-
     #[css(skip)]
     End, // Just for array-indexing purposes.
 }
@@ -393,28 +436,18 @@ pub enum SystemColor {
 #[cfg(feature = "gecko")]
 impl SystemColor {
     #[inline]
-    fn compute(&self, cx: &Context, scheme: SystemColorScheme) -> ComputedColor {
+    fn compute(&self, cx: &Context) -> ComputedColor {
+        use crate::gecko::values::convert_nscolor_to_rgba;
         use crate::gecko_bindings::bindings;
 
-        let prefs = cx.device().pref_sheet_prefs();
-
-        convert_nscolor_to_computedcolor(match *self {
-            SystemColor::Canvastext => prefs.mDefaultColor,
-            SystemColor::Canvas => prefs.mDefaultBackgroundColor,
-            SystemColor::Linktext => prefs.mLinkColor,
-            SystemColor::Activetext => prefs.mActiveLinkColor,
-            SystemColor::Visitedtext => prefs.mVisitedLinkColor,
-
-            _ => {
-                let color = unsafe {
-                    bindings::Gecko_GetLookAndFeelSystemColor(*self as i32, cx.device().document(), scheme)
-                };
-                if color == bindings::NS_SAME_AS_FOREGROUND_COLOR {
-                    return ComputedColor::currentcolor();
-                }
-                color
-            }
-        })
+        // TODO: We should avoid cloning here most likely, though it's
+        // cheap-ish.
+        let style_color_scheme = cx.style().get_inherited_ui().clone_color_scheme();
+        let color = cx.device().system_nscolor(*self, &style_color_scheme);
+        if color == bindings::NS_SAME_AS_FOREGROUND_COLOR {
+            return ComputedColor::currentcolor();
+        }
+        ComputedColor::rgba(convert_nscolor_to_rgba(color))
     }
 }
 
@@ -488,20 +521,6 @@ impl<'a, 'b: 'a, 'i: 'a> ::cssparser::ColorComponentParser<'i> for ColorComponen
     }
 }
 
-fn parse_moz_system_color<'i, 't>(
-    context: &ParserContext,
-    input: &mut Parser<'i, 't>,
-) -> Result<(SystemColor, SystemColorScheme), ParseError<'i>> {
-    debug_assert!(context.chrome_rules_enabled());
-    input.expect_function_matching("-moz-system-color")?;
-    input.parse_nested_block(|input| {
-        let color = SystemColor::parse(context, input)?;
-        input.expect_comma()?;
-        let scheme = SystemColorScheme::parse(input)?;
-        Ok((color, scheme))
-    })
-}
-
 impl Parse for Color {
     fn parse<'i, 't>(
         context: &ParserContext,
@@ -527,13 +546,7 @@ impl Parse for Color {
                 #[cfg(feature = "gecko")]
                 {
                     if let Ok(system) = input.try_parse(|i| SystemColor::parse(context, i)) {
-                        return Ok(Color::System(system, SystemColorScheme::Default));
-                    }
-
-                    if context.chrome_rules_enabled() {
-                        if let Ok((color, scheme)) = input.try_parse(|i| parse_moz_system_color(context, i)) {
-                            return Ok(Color::System(color, scheme));
-                        }
+                        return Ok(Color::System(system));
                     }
                 }
 
@@ -572,17 +585,7 @@ impl ToCss for Color {
             Color::Complex(_) => Ok(()),
             Color::ColorMix(ref mix) => mix.to_css(dest),
             #[cfg(feature = "gecko")]
-            Color::System(system, scheme) => {
-                if scheme == SystemColorScheme::Default {
-                    system.to_css(dest)
-                } else {
-                    dest.write_str("-moz-system-color(")?;
-                    system.to_css(dest)?;
-                    dest.write_str(", ")?;
-                    scheme.to_css(dest)?;
-                    dest.write_char(')')
-                }
-            }
+            Color::System(system) => system.to_css(dest),
             #[cfg(feature = "gecko")]
             Color::InheritFromBodyQuirk => Ok(()),
         }
@@ -711,12 +714,6 @@ impl Color {
     }
 }
 
-#[cfg(feature = "gecko")]
-fn convert_nscolor_to_computedcolor(color: nscolor) -> ComputedColor {
-    use crate::gecko::values::convert_nscolor_to_rgba;
-    ComputedColor::rgba(convert_nscolor_to_rgba(color))
-}
-
 impl Color {
     /// Converts this Color into a ComputedColor.
     ///
@@ -734,14 +731,16 @@ impl Color {
                 let left = mix.left.to_computed_color(context)?.to_animated_value();
                 let right = mix.right.to_computed_color(context)?.to_animated_value();
                 ToAnimatedValue::from_animated_value(AnimatedColor::mix(
+                    mix.color_space,
                     &left,
                     mix.left_percentage.get(),
                     &right,
                     mix.right_percentage.get(),
+                    mix.hue_adjuster,
                 ))
             },
             #[cfg(feature = "gecko")]
-            Color::System(system, scheme) => system.compute(context?, scheme),
+            Color::System(system) => system.compute(context?),
             #[cfg(feature = "gecko")]
             Color::InheritFromBodyQuirk => ComputedColor::rgba(context?.device().body_text_color()),
         })
@@ -807,7 +806,7 @@ impl SpecifiedValueInfo for Color {
         // should probably be handled that way as well.
         // XXX `currentColor` should really be `currentcolor`. But let's
         // keep it consistent with the old system for now.
-        f(&["rgb", "rgba", "hsl", "hsla", "currentColor", "transparent"]);
+        f(&["rgb", "rgba", "hsl", "hsla", "hwb", "currentColor", "transparent"]);
     }
 }
 
@@ -856,3 +855,141 @@ impl Parse for CaretColor {
         ColorOrAuto::parse(context, input).map(GenericCaretColor)
     }
 }
+
+bitflags! {
+    /// Various flags to represent the color-scheme property in an efficient
+    /// way.
+    #[derive(Default, MallocSizeOf, SpecifiedValueInfo, ToComputedValue, ToResolvedValue, ToShmem)]
+    #[repr(C)]
+    #[value_info(other_values = "light,dark,only")]
+    pub struct ColorSchemeFlags: u8 {
+        /// Whether the author specified `light`.
+        const LIGHT = 1 << 0;
+        /// Whether the author specified `dark`.
+        const DARK = 1 << 1;
+        /// Whether the author specified `only`.
+        const ONLY = 1 << 2;
+    }
+}
+
+/// <https://drafts.csswg.org/css-color-adjust/#color-scheme-prop>
+#[derive(
+    Clone,
+    Debug,
+    Default,
+    MallocSizeOf,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(C)]
+#[value_info(other_values = "normal")]
+pub struct ColorScheme {
+    #[ignore_malloc_size_of = "Arc"]
+    idents: crate::ArcSlice<CustomIdent>,
+    bits: ColorSchemeFlags,
+}
+
+impl ColorScheme {
+    /// Returns the `normal` value.
+    pub fn normal() -> Self {
+        Self {
+            idents: Default::default(),
+            bits: ColorSchemeFlags::empty(),
+        }
+    }
+
+    /// Returns the raw bitfield.
+    pub fn raw_bits(&self) -> u8 {
+        self.bits.bits
+    }
+}
+
+impl Parse for ColorScheme {
+    fn parse<'i, 't>(
+        _: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        let mut idents = vec![];
+        let mut bits = ColorSchemeFlags::empty();
+
+        let mut location = input.current_source_location();
+        while let Ok(ident) = input.try_parse(|i| i.expect_ident_cloned()) {
+            let mut is_only = false;
+            match_ignore_ascii_case! { &ident,
+                "normal" => {
+                    if idents.is_empty() && bits.is_empty() {
+                        return Ok(Self::normal());
+                    }
+                    return Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+                },
+                "light" => bits.insert(ColorSchemeFlags::LIGHT),
+                "dark" => bits.insert(ColorSchemeFlags::DARK),
+                "only" => {
+                    if bits.intersects(ColorSchemeFlags::ONLY) {
+                        return Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+                    }
+                    bits.insert(ColorSchemeFlags::ONLY);
+                    is_only = true;
+                },
+                _ => {},
+            };
+
+            if is_only {
+                if !idents.is_empty() {
+                    // Only is allowed either at the beginning or at the end,
+                    // but not in the middle.
+                    break;
+                }
+            } else {
+                idents.push(CustomIdent::from_ident(location, &ident, &[])?);
+            }
+            location = input.current_source_location();
+        }
+
+        if idents.is_empty() {
+            return Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+        }
+
+        Ok(Self {
+            idents: crate::ArcSlice::from_iter(idents.into_iter()),
+            bits,
+        })
+    }
+}
+
+impl ToCss for ColorScheme {
+    fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result
+    where
+        W: Write,
+    {
+        if self.idents.is_empty() {
+            debug_assert!(self.bits.is_empty());
+            return dest.write_str("normal");
+        }
+        let mut first = true;
+        for ident in self.idents.iter() {
+            if !first {
+                dest.write_char(' ')?;
+            }
+            first = false;
+            ident.to_css(dest)?;
+        }
+        if self.bits.intersects(ColorSchemeFlags::ONLY) {
+            dest.write_str(" only")?;
+        }
+        Ok(())
+    }
+}
+
+/// https://drafts.csswg.org/css-color-adjust/#print-color-adjust
+#[derive(Clone, Copy, Debug, MallocSizeOf, Parse, PartialEq, SpecifiedValueInfo, ToCss, ToComputedValue, ToResolvedValue, ToShmem)]
+#[repr(u8)]
+pub enum PrintColorAdjust {
+    /// Ignore backgrounds and darken text.
+    Economy,
+    /// Respect specified colors.
+    Exact,
+}
